%{
    #include "du3.h"
    #include "duerr.h"
    #include <assert.h>

    /* semantic value */
    #define YYSTYPE mlc::MlaskalLval

	/* locations */
	#define YYLTYPE unsigned
	#define YYLLOC_DEFAULT(cur, rhs, n)	do { if(n) (cur)=(rhs)[1]; else (cur)=(rhs)[0]; } while(0)

    /* local stuff */
    using namespace mlc;
    
	/* error stuff */
    #define YYERROR_VERBOSE 1
    static void yyerror(YYLTYPE *line, mlc::MlaskalCtx *ctx, const char *m);
    
%}

%token DUTOK_PROGRAM			    /* program */
%token DUTOK_LABEL			    /* label */
%token DUTOK_CONST			    /* const */
%token DUTOK_TYPE			    /* type */
%token DUTOK_VAR			    /* var */
%token DUTOK_BEGIN			    /* begin */
%token DUTOK_END			    /* end */
%token DUTOK_PROCEDURE			    /* procedure */
%token DUTOK_FUNCTION			    /* function */
%token DUTOK_ARRAY			    /* array */
%token DUTOK_OF				    /* of */
%token DUTOK_GOTO			    /* goto */
%token DUTOK_IF				    /* if */
%token DUTOK_THEN			    /* then */
%token DUTOK_ELSE			    /* else */
%token DUTOK_WHILE			    /* while */
%token DUTOK_DO				    /* do */
%token DUTOK_REPEAT			    /* repeat */
%token DUTOK_UNTIL			    /* until */
%token DUTOK_FOR			    /* for */
%token DUTOK_OR				    /* or */
%token DUTOK_NOT			    /* not */
%token DUTOK_RECORD			    /* record */

/* literals */
%token DUTOK_IDENTIFIER			    /* identifier */
%token DUTOK_UINT			    /* unsigned integer */
%token DUTOK_REAL			    /* real number */
%token DUTOK_STRING			    /* string */

/* delimiters */
%token DUTOK_SEMICOLON			    /* ; */
%token DUTOK_DOT			    /* . */
%token DUTOK_COMMA			    /* , */
%token DUTOK_EQ				    /* = */
%token DUTOK_COLON			    /* : */
%token DUTOK_LPAR			    /* ( */
%token DUTOK_RPAR			    /* ) */
%token DUTOK_DOTDOT			    /* .. */
%token DUTOK_LSBRA			    /* [ */
%token DUTOK_RSBRA			    /* ] */
%token DUTOK_ASSIGN			    /* := */

/* grouped operators and keywords */
%token DUTOK_OPER_REL			    /* <, <=, <>, >=, > */
%token DUTOK_OPER_SIGNADD		    /* +, - */
%token DUTOK_OPER_MUL			    /* *, /, div, mod, and */
%token DUTOK_FOR_DIRECTION		    /* to, downto */

%start mlaskal
%error-verbose
%pure-parser
%parse-param {mlc::MlaskalCtx *ctx}
%locations

%%

mlaskal:	  DUTOK_PROGRAM DUTOK_IDENTIFIER DUTOK_SEMICOLON program_block DUTOK_DOT
		;

program_block:	  label_block const_block type_block var_block method_block DUTOK_BEGIN stmt_seq DUTOK_END
		;

label_block:	  DUTOK_LABEL uint_seq DUTOK_SEMICOLON
		| /*no label block */
		;

const_block:	  DUTOK_CONST const_seq
		| /*no const block */
		;

type_block:	  DUTOK_TYPE type_seq
		| /*no type block */
		;

var_block:	  DUTOK_VAR var_seq
		| /*no var block */
		;

method_block:	  procedure_header DUTOK_SEMICOLON block DUTOK_SEMICOLON method_block
		| function_header DUTOK_SEMICOLON block DUTOK_SEMICOLON method_block
		| /*no method block */
		;

block:		  label_block const_block type_block var_block DUTOK_BEGIN stmt_seq DUTOK_END
			;

procedure_header: DUTOK_PROCEDURE DUTOK_IDENTIFIER DUTOK_LPAR formal_parametrs DUTOK_RPAR
		| DUTOK_PROCEDURE DUTOK_IDENTIFIER
		;

function_header:  DUTOK_FUNCTION DUTOK_IDENTIFIER DUTOK_LPAR formal_parametrs DUTOK_RPAR DUTOK_COLON DUTOK_IDENTIFIER /* scalar type identifier */
		| DUTOK_FUNCTION DUTOK_IDENTIFIER DUTOK_COLON DUTOK_IDENTIFIER /* scalar type identifier */
		;

formal_parametrs: var_optional identifier_seq DUTOK_COLON DUTOK_IDENTIFIER /* identifier of type */
		| var_optional identifier_seq DUTOK_COLON DUTOK_IDENTIFIER DUTOK_SEMICOLON formal_parametrs
		;

real_params:	expr
		| expr DUTOK_COMMA real_params
		;

/* various sequences */

uint_seq:	  DUTOK_UINT
		| DUTOK_UINT DUTOK_COMMA uint_seq
		;

stmt_seq:	  stmt
		| stmt DUTOK_SEMICOLON stmt_seq
		;

matched_seq:	  matched
		| matched DUTOK_SEMICOLON matched_seq
		;

unmatched_seq:	  unmatched
		| unmatched DUTOK_SEMICOLON unmatched_seq
		;

const_seq:	  DUTOK_IDENTIFIER DUTOK_EQ const_val DUTOK_SEMICOLON
		| DUTOK_IDENTIFIER DUTOK_EQ const_val DUTOK_SEMICOLON const_seq
		;

type_seq:	  DUTOK_IDENTIFIER DUTOK_EQ type_val DUTOK_SEMICOLON
		| DUTOK_IDENTIFIER DUTOK_EQ type_val DUTOK_SEMICOLON type_seq
		;

var_seq:	  identifier_seq DUTOK_COLON type_val DUTOK_SEMICOLON
		| identifier_seq DUTOK_COLON type_val DUTOK_SEMICOLON var_seq
		;

identifier_seq:	  DUTOK_IDENTIFIER
		| DUTOK_IDENTIFIER DUTOK_COMMA identifier_seq
		;

ord_type_seq:	  ord_type_id
		| ord_type_seq DUTOK_COMMA ord_type_id
		;

ordinal_expr_seq: expr /* expr of ordinal type */
		| expr DUTOK_COMMA ordinal_expr_seq
		;

term_seq:	  term
		| term DUTOK_OPER_SIGNADD term_seq
		| term DUTOK_OR term_seq
		;

/* various values */

const_val:	  const_nosign
		| DUTOK_IDENTIFIER /* of konst */ 
		| DUTOK_OPER_SIGNADD DUTOK_UINT
		| DUTOK_OPER_SIGNADD DUTOK_REAL
		;

const_nosign:	  DUTOK_UINT
		| DUTOK_REAL
		| DUTOK_STRING
		;

type_val:	  ord_type_noid
		| struc_type_val
		| DUTOK_IDENTIFIER
		| DUTOK_IDENTIFIER DUTOK_DOTDOT ord_const_id
		;

ord_type_id:	  ord_const_noid DUTOK_DOTDOT ord_const_id
		| DUTOK_IDENTIFIER DUTOK_DOTDOT ord_const_id
		| DUTOK_IDENTIFIER
		;

ord_type_noid:	  ord_const_noid DUTOK_DOTDOT ord_const_id
		;

struc_type_val:	  DUTOK_ARRAY DUTOK_LSBRA ord_type_seq DUTOK_RSBRA DUTOK_OF type_val
		;

ord_const_noid:	  DUTOK_OPER_SIGNADD DUTOK_IDENTIFIER /* identifier of const integer */
		| sign_add_opt DUTOK_UINT
		;

ord_const_id:	  sign_add_opt DUTOK_IDENTIFIER /* identifier of const integer */
		| sign_add_opt DUTOK_UINT
		;

stmt:		  matched
		| unmatched 
		;

matched:	  label_inst_opt DUTOK_IF expr DUTOK_THEN matched DUTOK_ELSE matched
		| label_inst_opt variable DUTOK_ASSIGN expr
		| label_inst_opt DUTOK_IDENTIFIER /* identifier of function or variable */ DUTOK_ASSIGN expr
		| label_inst_opt DUTOK_IDENTIFIER /* identifier of procedure */ 
		| label_inst_opt DUTOK_IDENTIFIER /* identifier of procedure */ DUTOK_LPAR real_params DUTOK_RPAR
		| label_inst_opt DUTOK_GOTO DUTOK_UINT
		| label_inst_opt DUTOK_BEGIN stmt_seq DUTOK_END
		| label_inst_opt DUTOK_REPEAT matched_seq DUTOK_UNTIL expr /* expr of bool */ 
		| label_inst_opt DUTOK_FOR DUTOK_IDENTIFIER /* identifier of ordinal type variable */ DUTOK_ASSIGN expr /* ordinal type expr */ DUTOK_FOR_DIRECTION expr /* expr of ordinal */ DUTOK_DO matched
		| label_inst_opt DUTOK_WHILE expr /* expr of bool */ DUTOK_DO matched 
		| 
		;
		
unmatched:	  label_inst_opt DUTOK_IF expr /* expr of bool */ DUTOK_THEN stmt
		| label_inst_opt DUTOK_IF expr /* expr of bool */ DUTOK_THEN matched DUTOK_ELSE unmatched
		| label_inst_opt DUTOK_FOR DUTOK_IDENTIFIER /* identifier of ordinal type variable */ DUTOK_ASSIGN expr /* ordinal type expr */ DUTOK_FOR_DIRECTION expr /* expr of ordinal */ DUTOK_DO unmatched /* ?? matched ?? */
		| label_inst_opt DUTOK_WHILE expr /* expr of bool */ DUTOK_DO unmatched 
		| label_inst_opt DUTOK_REPEAT unmatched_seq DUTOK_UNTIL expr /* expr of bool */ 
		;

expr:		  simple_expr
		| simple_expr DUTOK_OPER_REL simple_expr
		| simple_expr DUTOK_EQ simple_expr
		;

simple_expr:	  term_seq
		| DUTOK_OPER_SIGNADD term_seq
		;

sign_add_opt:	  DUTOK_OPER_SIGNADD
		| /* no sign add */
		;

term:		  faktor
		| term DUTOK_OPER_MUL faktor
		;

faktor:		  const_nosign
		| variable
		| DUTOK_IDENTIFIER /* identifier of variable or const or ... */
		| DUTOK_IDENTIFIER /*identifier of function */ DUTOK_LPAR real_params DUTOK_RPAR
		| DUTOK_LPAR expr DUTOK_RPAR
		| DUTOK_NOT faktor
		;

variable:	  DUTOK_IDENTIFIER /* identifier of array variable */ DUTOK_LSBRA ordinal_expr_seq DUTOK_RSBRA
		| variable /* identifier of array variable */ DUTOK_LSBRA ordinal_expr_seq DUTOK_RSBRA
		;

var_optional:	  DUTOK_VAR
		| /* no var keyword */
		;

label_inst_opt:	  DUTOK_UINT DUTOK_COLON
		| /* no label */
		;

%%

static void yyerror(unsigned *line, MlaskalCtx *, const char *m)
{
    error(DUERR_SYNTAX, *line, m);
}
